// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   Grayscale Image
 * 		Implementation of the Primitive Operations
 *
 *   The image is represented as a Pixmap, i.e. a matrix of pixels
 *    each of them specifies the gray level at a particular point
 *
 * $Id: image.cc,v 3.0 2000/10/25 02:06:39 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifdef __GNUC__
#pragma implementation "image.h"
#endif

#include "image.h"
#include "std.h"
#include "minmax.h"
#include <math.h>
#include "iostream.h"



/*
 *------------------------------------------------------------------------
 *			Constructors and destructors
 */

void IMAGE::allocate(
	const card no_rows,		// No. of rows
	const card no_cols,		// No. of cols
	const card depth		// No. of bits per pixel
)
{
  valid_code = IMAGE_val_code;

  assure((ncols=no_cols) > 0, "Zero image width unexpected");
  assure((nrows=no_rows) > 0, "Zero image height unexpected");
  assure((bits_per_pixel=depth) > 0 && depth <= GRAY_MAXBIT, 
	 "Zero or too large no. of bits per pixel");

  name    = "";
  npixels = nrows * ncols;

  assert( (scanrows = (GRAY **)calloc(nrows,sizeof(GRAY *))) != 0 );
  assert( (pixels   = (GRAY *)calloc(npixels,sizeof(GRAY))) != 0 );

  register GRAY * pp = pixels;
  register GRAY **sp = scanrows;
  while( pp < pixels + npixels )	// Build the row index
    *sp++ = pp, pp += ncols;
  assert( sp == scanrows + nrows );
}

				// Set a new image name
void IMAGE::set_name(const char * new_name)
{
  if( name != 0 && name[0] != '\0' )	// Dispose of the previous image name
    delete const_cast<char*>(name);

  if( new_name == 0 || new_name[0] == '\0' )
    name = "";				// Image is anonymous now
  else
    name = new char[strlen(new_name)+1],
      strcpy(const_cast<char *>(name),new_name);
}


IMAGE::~IMAGE(void)		// Dispose the image struct
{
  is_valid();
  if( name != 0 && name[0] != '\0' )
    delete const_cast<char*>(name);

  free(scanrows);
  free(pixels);
  valid_code = 0;
}

/*
 *------------------------------------------------------------------------
 *	    	    Expansion/Shrinking of the image
 */

				// Expand the prototype twice in each
				// dimension
				// In other words, blow each pixel of the
				// prototype to the 2x2 square
void LazyExpandedImage::fill_in(IMAGE& im) const
{
  register GRAY * tp = im.pixels;
  register const GRAY * pp = proto.pixels;

  for(register card i=0; i<proto.nrows; i++, tp+=im.ncols)
    for(register card j=0; j<proto.ncols; j++)
    {
      register GRAY pixel = *pp++;
      *(tp+ncols) = pixel;		// Fill out 2x2 square of the
      *tp++       = pixel;		// blown image with a pixel value
      *(tp+ncols) = pixel;
      *tp++       = pixel;
    }

  assert( pp == proto.pixels + proto.npixels );
  assert( tp == im.pixels + im.npixels );
}

				// Make an image half the size of the
				// proto image
				// The image row and column
				// dimensions are assumed to be even numbers
const LazyShrunkImage shrink_twice(const IMAGE & proto)
{  proto.is_valid(); 
    if( (proto.q_nrows() & 1) || (proto.q_ncols() & 1) )
    _error("No of rows and columns in the image to shrink should both be even",
	   (proto.info(),0));
    return proto; 
}

				// Image is shrunk by replacing a 2x2 square
				// by one pixel with an average intensity
				// over the square
void LazyShrunkImage::fill_in(IMAGE& im) const
{
  register GRAY * tp = im.pixels;
  register const GRAY * pp = proto.pixels;

  for(register card i=0; i<im.nrows; i++)
  {					// Average row-by-row
    register card j;
    for(j=0; j<im.ncols; j++)
      *tp = *pp++, *tp++ += *pp++;	// Sum of pairs of the prototype pixels
    tp -= ncols;
    for(j=0; j<im.ncols; j++, tp++, pp+=2)
    {
      register int av = *tp + pp[0] + pp[1];	// This is the aver of 4 pixels
      *tp = ( av & 2 ? (av >> 2) + 1 : av >> 2 );	// /4 with rounding
    }
  }
  assert( pp == proto.pixels + proto.npixels );
  assert( tp == im.pixels + im.npixels );
}


				// Assign another to '*this' resizing
				// '*this' as necessary to fit exactly
				// Squeezing is done by subsampling (dropping
				// some image pixels), while stretching is done
				// by repeating pixels/scanlines
				// Note, the ratio of sizes of the images
				// can be *anything*
IMAGE& IMAGE::coerce(const IMAGE& another)
{
					// Make the width even
//  width  = ( dest_rect.q_width() + 1 ) & (~1);
//  height = dest_rect.q_height();
//  assert( width > 0 && height > 0 );

					// Figure out how much to
					// sqeeze/stretch
  const float ratio_h = (float)ncols/another.ncols;
  const float ratio_v = (float)nrows/another.nrows;

  float curr_v = 0, curr_h;		// Precise curr. loc. in window coord
  card dest_y = 0, dest_x;		// approx (rounded-off) coordinates
  register GRAY * drp = pixels;		// Scanline ptr for fitted image
  register const GRAY * spp = another.pixels; // Scanline of 'another' image

 					// Note, dest_x,dest_y is incremented
					// by (inc_v,inc_h), but curr_v is
					// incremented by ratio_v, etc.
					// Draw horizontally (row-by-row)
  while( spp < another.pixels + another.npixels )
  {
    curr_v += ratio_v;
    const GRAY * const spp_end = spp + another.ncols; // end of scanline

    int inc_v = (int)curr_v - dest_y;
    if( inc_v == 0 )			// inc_v = 0 means this row of 
    {					// orig_image is to be skipped
      spp = spp_end;
      continue;
    }
    dest_y = (int)curr_v;

    curr_h = 0;
    dest_x = 0;
    register GRAY * dpp = drp;		// Squeeze/stretch pixels in a scanline
    for(; spp < spp_end; spp++)
    {
      curr_h += ratio_h;

      int inc_h = (int)curr_h - dest_x;
      dest_x = (int)curr_h;
      while( inc_h-- > 0 )
	*dpp++ = *spp;
    }
    if( dpp == drp+ncols-1 )		// Can happen due to roundoff arithm
      *dpp++ = *(spp-1);
    assert( dpp == drp+ncols );
					// Stretching vertically involves
					// duplicating scanlines
    while( --inc_v > 0 )
      memcpy((void *)(drp+ncols),(void *)drp,ncols*sizeof(GRAY)),
      drp += ncols;
    drp += ncols;
  }
  if( drp == pixels + npixels-ncols )	// Duplicate the last scanline
    memcpy((void *)drp,(void *)(drp-ncols),ncols*sizeof(GRAY)), // in case
    drp += ncols;			// of roundoff snag

  assert( drp == pixels + npixels );

  return *this;
}

/*
 *------------------------------------------------------------------------
 * 			Collection-scalar operations
 *	walk through the collection and do something with each
 *		    (signed)pixel in turn, and the scalar
 */

				// For every element, do `pixel OP value`
#define COMPUTED_VAL_ASSIGNMENT(OP,VALTYPE)				\
									\
void PixelWise::operator OP (const VALTYPE val)			\
{									\
  register GRAY * cp = start_ptr;					\
  while( cp < end_ptr )							\
    *cp++ OP val;							\
}									\

COMPUTED_VAL_ASSIGNMENT(=,GRAY)
COMPUTED_VAL_ASSIGNMENT(+=,int)
COMPUTED_VAL_ASSIGNMENT(-=,int)
COMPUTED_VAL_ASSIGNMENT(*=,int)
COMPUTED_VAL_ASSIGNMENT(|=,int)
COMPUTED_VAL_ASSIGNMENT(&=,int)
COMPUTED_VAL_ASSIGNMENT(^=,int)

#undef COMPUTED_VAL_ASSIGNMENT

				// is "(signed)pixel OP val" true for all
				// pixels?

#define COMPARISON_WITH_SCALAR(OP)					\
									\
bool PixelWiseConst::operator OP (const int val) const			\
{									\
  register const GRAY * cp = start_ptr;					\
  while( cp < end_ptr )							\
    if( !(static_cast<GRAY_SIGNED>(*cp++) OP val) )			\
      return false;							\
									\
  return true;								\
}									\

COMPARISON_WITH_SCALAR(!=)
COMPARISON_WITH_SCALAR(<)
COMPARISON_WITH_SCALAR(<=)
COMPARISON_WITH_SCALAR(>)
COMPARISON_WITH_SCALAR(>=)

#undef COMPARISON_WITH_SCALAR


bool PixelWiseConst::operator == (const GRAY val) const
{
  register const GRAY * cp = start_ptr;
  while( cp < end_ptr )
    if( *cp++ != val )
      return false;
  return true;
}

				// Shift the value of all the pixels
void PixelWise::operator <<= (const card val)
{
  if( val >= GRAY_MAXBIT )
    _error("Very fishy shift factor: %d",val);

  register GRAY * cp = start_ptr;
  for(; cp < end_ptr; cp++)
    *cp <<= val;
}

				// Shift the value of all the pixels
void PixelWise::operator >>= (const card val)
{
  if( val >= GRAY_MAXBIT )
    _error("Very fishy shift factor: %d",val);

  register GRAY * cp = start_ptr;
  for(; cp < end_ptr; cp++)
    *cp = static_cast<GRAY_SIGNED>(*cp) >> val;
}

/*
 *------------------------------------------------------------------------
 *	Apply algebraic and other functions to all pixels of a collection
 */

				// Clip pixel values to
				// [0,1<<bits_per_pixel-1].
				// A pixel with the value outside that range
				// (i.e. negative or too big) is set to
				// 0 or 1<<bits_per_pixel-1, resp.
void PixelWise::clip_to_intensity_range(void)
{
  const int maxval = (1<<bits_per_pixel)-1;
  register GRAY * cp = start_ptr;
  for(; cp < end_ptr; cp++ )
    if( static_cast<GRAY_SIGNED>(*cp) < 0 )
      *cp = 0;
    else if( static_cast<GRAY_SIGNED>(*cp) > maxval )
      *cp = maxval;
}

				// Perform a transformation
				// pixel = |(signed)pixel|
				// on all the pixels of the image
void PixelWise::abs(void)
{
  register GRAY * cp = start_ptr;
  for(; cp < end_ptr; cp++ )
    if( static_cast<GRAY_SIGNED>(*cp) < 0 )
      *cp = -(*cp);
}

void PixelWise::clear(void)	// Clear the pixels in the collection
{
  memset(start_ptr,0,sizeof(GRAY)*(end_ptr-start_ptr));
}

void PixelWise::invert(void)	// Invert the pixels in the collection
{
  *this ^= ((1<<bits_per_pixel)-1);
}

IMAGE& IMAGE::invert(void)
{
  to_every(*this).invert();
  return *this;
}


				// Apply a user-defined action to each pixel
				// while image is efficiently traversed
				// row-by-row
IMAGE& IMAGE::apply(const PixelAction& action)
{
  is_valid();
  register GRAY * cp = pixels;
  for(register card i=0; i<nrows; i++)
    for(register card j=0; j<ncols; j++)
      const_cast<PixelAction&>(action).operation(*cp++,i,j);
  assert( cp == pixels+npixels );

  return *this;
}

const IMAGE& IMAGE::apply(const ConstPixelAction& action) const
{
  is_valid();
  register GRAY * cp = pixels;
  for(register card i=0; i<nrows; i++)
    for(register card j=0; j<ncols; j++)
      const_cast<ConstPixelAction&>(action).operation(*cp++,i,j);
  assert( cp == pixels+npixels );

  return *this;
}

				// Perform the histogram equalization
void PixelWise::equalize(const int no_grays)
{
  const int orig_no_grays = 1 << bits_per_pixel;
  assert( no_grays <= orig_no_grays );

  int * const histogram = (int *)calloc(orig_no_grays,sizeof(int));

				// Evaluate the histogram of an image
  register GRAY * cp = start_ptr;
  for(; cp < end_ptr; cp++ )
  {
    if( static_cast<GRAY_SIGNED>(*cp) <= 0 )
      *cp = 0;
    else if( static_cast<GRAY_SIGNED>(*cp) >= orig_no_grays )
      *cp = orig_no_grays-1;
    histogram[*cp]++;
  }

  const int pixels_per_bin_optimal = 
    ((end_ptr - start_ptr) + no_grays - 1)/no_grays;
  const int gray_shade_subsample_factor = 
    (orig_no_grays + no_grays -1)/no_grays;
  register int pixel;
  int new_pixel, new_pixel_old;
  int accumulated = 0;
  int optimal_accumulated = pixels_per_bin_optimal;
					// Mapping from pixels of original
  					// image to [new_pixel_old,new_pixel]
  short * const look_up_table = (short *)malloc(orig_no_grays*sizeof(short));
					// (actually, just to the center of
					// this interval)

  					// Equalizing the histogram
  for(pixel=0,new_pixel=0,new_pixel_old=0; pixel<orig_no_grays; pixel++)
  {
    accumulated += histogram[pixel];
    while( accumulated > optimal_accumulated )
      new_pixel += gray_shade_subsample_factor, 
      optimal_accumulated += pixels_per_bin_optimal;
    assert( new_pixel < orig_no_grays );
    look_up_table[pixel] = (new_pixel > 
			    new_pixel_old+gray_shade_subsample_factor ?
			    (new_pixel + new_pixel_old)/2 :
			    new_pixel);
    new_pixel_old = new_pixel;
  }

				// Update the image according to the LUT
  for(cp = start_ptr; cp < end_ptr; cp++)
    *cp = look_up_table[static_cast<GRAY_SIGNED>(*cp)];

  free(histogram);
  free(look_up_table);
}

/*
 *------------------------------------------------------------------------
 * 		Pixel-wise operations on two groups of pixels
 */

				// Check to see if two images are identical
bool IMAGE::operator == (const IMAGE& im) const
{
  are_compatible(*this,im);
  return (memcmp(pixels,im.pixels,npixels*sizeof(GRAY)) == 0);
}

bool PixelWiseConst::operator == (const PixelWiseConst& another) const
{
  sure_compatible_with(another);
  register const GRAY * sp = another.start_ptr;
  register const GRAY * tp = start_ptr;
  while( tp < end_ptr )
    if( *tp++ != *sp++ )
      return false;
  return true;
}

				// For every pixel, do `pixel OP another.pixel`
#define TWO_GROUP_COMP(OP)						\
									\
bool PixelWiseConst::operator OP (const PixelWiseConst& another) const  \
{									\
  sure_compatible_with(another);					\
  register const GRAY * sp = another.start_ptr;				\
  register const GRAY * tp = start_ptr;					\
  while( tp < end_ptr )							\
    if( !(static_cast<GRAY_SIGNED>(*tp++) OP static_cast<GRAY_SIGNED>(*sp++)) )						\
      return false;							\
									\
  return true;								\
}									\

TWO_GROUP_COMP(!=)
TWO_GROUP_COMP(<)
TWO_GROUP_COMP(<=)
TWO_GROUP_COMP(>)
TWO_GROUP_COMP(>=)

#undef TWO_GROUP_COMP


IMAGE& IMAGE::operator = (const IMAGE& source)
{
  are_compatible(*this,source);
  memcpy(pixels,source.pixels,npixels*sizeof(GRAY));
  return *this;
}

				// For every pixel, do `pixel OP another.pixel`
#define TWO_GROUP_OP(OP)						\
									\
void PixelWise::operator OP (const PixelWiseConst& another)		\
{									\
  sure_compatible_with(another);					\
  register const GRAY * sp = another.start_ptr;				\
  register GRAY * tp = start_ptr;					\
  while( tp < end_ptr )							\
    *tp++ OP *sp++;							\
}									\

TWO_GROUP_OP(=)
TWO_GROUP_OP(+=)
TWO_GROUP_OP(-=)
TWO_GROUP_OP(|=)
TWO_GROUP_OP(&=)
TWO_GROUP_OP(^=)

#undef TWO_GROUP_OP

#if 0
				// Modified addition
				//	Target += scalar*Source
IMAGE& add(IMAGE& target, const int scalar,const IMAGE& source)
{
  are_compatible(target,source);

  register GRAY * sp = source.pixels;
  register GRAY * tp = target.pixels;
  while( tp < target.pixels + target.npixels )
    *tp++ += scalar * *sp++;
  
  return target;
}
				// Shift the source to 'pos', clip it 
				// if necessary, multiply by the scalar,
				// and add
IMAGE& IMAGE::shift_clip_add(rowcol pos, const int scalar, const IMAGE& source)
{
  source.is_valid();
  is_valid();

					// Find an intersection of 'source'
					// shifted to 'pos' (relative to
					// the image) with the (target) image
					// Only this rectangle will be 
					// affected by addition
  rowcol s_orig(max(0,-pos.row()),max(0,-pos.col()));
  rowcol t_orig(max(0,pos.row()),max(0,pos.col()));
					// No. of rows in the intersection
  const int irows = min(nrows - t_orig.row(),source.nrows - s_orig.row());
					// No. of cols in the intersection
  const int icols = min(ncols - t_orig.col(),source.ncols - s_orig.col());
  if( irows <= 0 || icols <= 0 )
    source.info(),
    info(),
    _error("The first image shifted by (%d,%d) does not intersect the other",
	   pos.row(),pos.col());

					// See the explanation in the class
					// Rectangle
  const int s_inc_to_nextrow = source.ncols - icols;	
  const int t_inc_to_nextrow = ncols - icols;	

  register GRAY_SIGNED * sp = 
    (GRAY_SIGNED *)&(source.scanrows[s_orig.row()])[s_orig.col()];
  register GRAY_SIGNED * tp = 
    (GRAY_SIGNED *)&(scanrows[t_orig.row()])[t_orig.col()];

  register int i,j;
  for(i=0; i<irows; i++, sp += s_inc_to_nextrow, tp += t_inc_to_nextrow)
    for(j=0; j<icols; j++)
      *tp++ += scalar * *sp++;

  return *this;
}
#endif

/*
 *------------------------------------------------------------------------
 *	Reduce a collection or a difference between two collections
 *		to a single number: a "norm"
 */

#define REDUCE_SUM(X,VAL) X += static_cast<GRAY_SIGNED>(VAL)
#define REDUCE_SUMSQ(X,VAL) X += sqr(static_cast<GRAY_SIGNED>(VAL))
#define REDUCE_SUMABS(X,VAL) X += ::abs(static_cast<GRAY_SIGNED>(VAL))
#define REDUCE_MAXABS(X,VAL) X = ::max(X,::abs(static_cast<GRAY_SIGNED>(VAL)))

#define REDUCE_SUMSQ2(X,VAL1,VAL2) X += sqr(static_cast<GRAY_SIGNED>(VAL1) - static_cast<GRAY_SIGNED>(VAL2))
#define REDUCE_SUMABS2(X,VAL1,VAL2) X += ::abs(static_cast<GRAY_SIGNED>(VAL1) - static_cast<GRAY_SIGNED>(VAL2))
#define REDUCE_MAXABS2(X,VAL1,VAL2) X = ::max(X,::abs(static_cast<GRAY_SIGNED>(VAL1) - static_cast<GRAY_SIGNED>(VAL2)))

#define REDUCE_ONE(RT,NAME,OP)						\
									\
RT PixelWiseConst::NAME (void) const					\
{									\
  register RT norm = 0;							\
  register GRAY * cp = start_ptr;					\
  while( cp < end_ptr )							\
    OP(norm,*cp++);							\
  return norm;								\
}									\



REDUCE_ONE(double,sum,REDUCE_SUM)
REDUCE_ONE(double,sum_squares,REDUCE_SUMSQ)
REDUCE_ONE(double,sum_abs,REDUCE_SUMABS)
REDUCE_ONE(int,max_abs,REDUCE_MAXABS)

#undef REDUCE_ONE

#define REDUCE_DIFF_OF_TWO(RT,NAME,OP)					\
									\
RT PixelWiseConst::NAME (const PixelWiseConst& another) const		\
{									\
  sure_compatible_with(another);					\
  register RT norm = 0;							\
  register const GRAY * sp = another.start_ptr;				\
  register const GRAY * tp = start_ptr;					\
  while( tp < end_ptr )							\
    OP(norm,*tp++,*sp++);						\
  return norm;								\
}									\

REDUCE_DIFF_OF_TWO(double,sum_squares,REDUCE_SUMSQ2)
REDUCE_DIFF_OF_TWO(double,sum_abs,REDUCE_SUMABS2)
REDUCE_DIFF_OF_TWO(int,max_abs,REDUCE_MAXABS2)

#undef REDUCE_DIFF_OF_TWO

#undef REDUCE_SUM
#undef REDUCE_SUMSQ
#undef REDUCE_SUMABS
#undef REDUCE_MAXABS
#undef REDUCE_SUMSQ2
#undef REDUCE_SUMABS2
#undef REDUCE_MAXABS2

/*
 *------------------------------------------------------------------------
 *			Compute pixel norms
 */

				// Compute the 1. norm of the entire image
				// SUM{ |(signed)pixel[i,j]| }
double norm_1(const IMAGE& im)
{
  return of_every(im).sum_abs();
}

				// Compute the square of the 2. norm of 
				// the entire image
				// SUM{ |(signed)pixel[i,j]|^2 }
double norm_2_sqr(const IMAGE& im)
{
  return of_every(im).sum_squares();
}

				// Compute the infinity norm of the
				// entire image
				// MAX{ |(signed)pixel[i,j]| }
int norm_inf(const IMAGE& im)
{
  return of_every(im).max_abs();
}

				// Find extremum values of image pixels
Extrema::Extrema(const IMAGE& image)
	: max_pixel(0,0), min_pixel(0,0)
{
  image.is_valid();
  max_value = min_value = image(0,0);

  register GRAY_SIGNED * cp = (GRAY_SIGNED *)image.pixels;
  for(register card i=0; i<image.nrows; i++)
    for(register card j=0; j<image.ncols; j++, cp++)
      if( *cp > max_value )
	max_value = *cp, max_pixel.row_val = i, max_pixel.col_val = j;
      else if( *cp < min_value )
	min_value = *cp, min_pixel.row_val = i, min_pixel.col_val = j;
}

				// Evaluate the scalar product of two
				// collections
double PixelWiseConst::operator * (const PixelWiseConst& another) const
{
  sure_compatible_with(another);

  register GRAY * p1 = start_ptr;
  register GRAY * p2 = another.start_ptr;
  register double sum = 0;

  while( p1 < end_ptr )
    sum += static_cast<GRAY_SIGNED>(*p1++) * static_cast<GRAY_SIGNED>(*p2++);

  return sum;
}

			// Estimate the norm of the difference 
			// between two (signed) image
				// SUM{ |(signed)pixel[i,j]| }
double norm_1(const IMAGE& im1, const IMAGE& im2)
{
  return of_every(im1).sum_abs(of_every(im2));
}

				// SUM{ |(signed)pixel[i,j]|^2 }
double norm_2_sqr(const IMAGE& im1, const IMAGE& im2)
{
  return of_every(im1).sum_squares(of_every(im2));
}
				// MAX{ |(signed)pixel[i,j]| }
int norm_inf(const IMAGE& im1, const IMAGE& im2)
{
  return of_every(im1).max_abs(of_every(im2));
}

				// Normalize pixel values to be
				// in range 0..1<<bits_per_pixel-1
void PixelWise::normalize_for_display(void)
{
  int minp = static_cast<GRAY_SIGNED>(*start_ptr);
  int maxp = minp;
  for(register const GRAY * cp = start_ptr; cp < end_ptr; cp++)
    if( static_cast<GRAY_SIGNED>(*cp) > maxp )
      maxp = static_cast<GRAY_SIGNED>(*cp);
    else if( static_cast<GRAY_SIGNED>(*cp) < minp )
      minp = static_cast<GRAY_SIGNED>(*cp);

  cerr<< "\nImage_normalization:\n\tmin pixel value " << minp
      << ", max pixel value " << maxp << endl;
  if( maxp != minp )
  {
    double factor = (double)((1<<bits_per_pixel)-1) /
		    ( maxp - minp );
    cerr << "\n\tNormalization is as follows: (pixel - " << minp
         << ")*" << factor << endl;
    for(register GRAY * cp = start_ptr; cp < end_ptr; cp++)
      *cp = static_cast<GRAY>( (static_cast<GRAY_SIGNED>(*cp) -  minp) * factor + 0.5 );
  }
}

/*
 *------------------------------------------------------------------------
 * 			Some service operations
 */

				// Print some info about the image
void IMAGE::info(void) const
{
  is_valid();
  cerr << "\nImage " << nrows << "x" << ncols << "x" << bits_per_pixel
       << " '" << name << "'" << endl;
}

  			// The image in question is not valid
  			// ==> crash the program
volatile void IMAGE::invalid_image(void) const
{
  _error("An operation attempted on an invalid image reference: 0x%0x\n",
  	(long)this);
}

				// Print the image as a table of pixel values
				// (zeros are printed as dots)
void IMAGE::print(const char * title) const
{
  is_valid();
  cerr << title << " follows " << endl;
  info();

  for(register card i=0; i<nrows; i++)
  {
    for(register card j=0; j<ncols; j++)
      if( (*this)(i,j) == 0 )
	message("   . ");
      else
	message("%4d ",static_cast<GRAY_SIGNED>((*this)(i,j)));
    message("\n");
  }
  message("Done\n");
}


void compare(			// Compare the two images
	const IMAGE& image1,	// and print out the result of comparison
	const IMAGE& image2,
	const char * title )
{
  are_compatible(image1,image2);

  cerr<< "\n\nComparison of two images:\n\t" << title;
  image1.info();
  image2.info();
 
  if( image1.bits_per_pixel != image2.bits_per_pixel )
    cerr << "\nNote, images have different depth, "
         << image1.bits_per_pixel << " and " << image2.bits_per_pixel
         << endl;

  double norm1 = 0, norm2 = 0;		// Norm of the images
  double ndiff = 0;			// Norm of the difference
  card imax=0,jmax=0;			// For the pixels that differ most
  int difmax = -1;			// For the pixels that differ most
					// Image scanline pointer
  register const GRAY *rowp1 = image1.pixels;
  register const GRAY *rowp2 = image2.pixels;

  bool im1_neg_pixel = false;		// Flags whether negative pixels
  bool im2_neg_pixel = false;		// were encountered

  for(register card i=0; i < image1.nrows; i++)
    for(register card j=0; j < image1.ncols; j++)
    {
      int pv1 = static_cast<GRAY_SIGNED>(*rowp1++);
      int pv2 = static_cast<GRAY_SIGNED>(*rowp2++);
      int diff = abs(pv1-pv2);

      if( pv1 < 0 )
	im1_neg_pixel = true, pv1 = -pv1;
      if( pv2 < 0 )
	im2_neg_pixel = true, pv2 = -pv2;

      if( diff > difmax )
      {
	difmax = diff;
	imax = i;
	jmax = j;
      }
      norm1 += pv1;
      norm2 += pv2;
      ndiff += diff;
    }

  if( im1_neg_pixel )
    cerr << "\n*** Warning: a pixel with negative value has been encountered "
	    "in image 1\n";
  if( im2_neg_pixel )
    cerr << "\n*** Warning: a pixel with negative value has been encountered "
	    "in image 2\n";

  cerr << "\nMaximal discrepancy    \t\t" << difmax;
  cerr << "\n   occured at the pixel\t\t(" << imax << ',' << jmax << ')';
  const int pv1 = image1(imax,jmax);
  const int pv2 = image2(imax,jmax);
  cerr << "\n Image 1 pixel is    \t\t" << pv1;
  cerr << "\n Image 2 pixel is    \t\t" << pv2;
  cerr << "\n Absolute error v2[i]-v1[i]\t\t" << pv2-pv1;
  cerr << "\n Relative error\t\t\t\t"
       << (pv2-pv1)/max((float)abs(pv2+pv1)/2,1e-7) << endl;

  cerr << "\nL1 norm of image 1 per pixel \t\t"
       << (double)norm1/image1.npixels;
  cerr << "\nL1 norm of image 2 per pixel \t\t"
       << (double)norm2/image2.npixels;
  cerr << "\nL1 norm of image1-image2 per pixel\t\t\t"
       << (double)ndiff/image1.npixels;
  cerr << "\n||Image1-Image2||/sqrt(||Image1|| ||Image2||)\t"
       << ndiff/max( sqrt(norm1*norm2), 1e-7 ) << endl << endl;
}


					// Verify the pixels have the value
					// that is expected
void verify_pixel_value(const IMAGE& im, const GRAY val)
{
  for(register card i=0; i<im.q_nrows(); i++)
    for(register card j=0; j<im.q_ncols(); j++)
      if( im(i,j) != val )
	_error("Pixel [%d,%d] has the value 0x%x different from expected 0x%x",
	       i,j,im(i,j),val);
}

